%% Core library module.
%%
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Julio Mariño Carballo
%%   * Jose María Rey Poza
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module('Corelib', []).

:- use_module(library(iso_byte_char), [char_code/2] ).
:- use_module(library(freeze),        [freeze/2]).

%% This file is an example of how to write a new prolog module for
%% Sloth. (It should be accompanied by their corresponding .curry
%% file, unless you want to write the type declarations by
%% hand, which is not recommended at all, as they can change.)

:- use_module('Core', [tonf/2, tohnf/2]).

:- multifile     user_tohnf/2, user_tonf/2.
:- discontiguous user_tohnf/2, user_tonf/2.

%% The block primitive:
user_tohnf('Prelude_block'(V, B)#R, R) :-
	!,
	tonf(V, VNF),
	freeze(VNF, tohnf(B, R)).

%% --- Converts a character into its ASCII value.
%% ord :: Char -> Int
%% ord external
user_tohnf('Prelude_ord'(Char)#R, R) :-
	!,
	tonf(Char, NFChar),
	freeze(NFChar, (
			 char_code(NFChar, R)
		     )).

%% --- Converts an ASCII value into a character.
%% chr :: Int -> Char
%% chr external
user_tohnf('Prelude_chr'(Int)#R, R) :-
	!, 
	tonf(Int, NFInt),
	freeze(NFInt, (
			  char_code(R, Int)
		      )).

%% Integer operations. They use freeze for simulation suspension.

%% --- Adds two integers.
%% (+)   :: Int -> Int -> Int
%% (+) external
user_tohnf('Prelude_+'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F + G),
		      freeze(B, true)
		  )).
%% --- Subtracts two integers.
%% (-)   :: Int -> Int -> Int
%% (-) external
user_tohnf('Prelude_-'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F - G),
		      freeze(B, true)
		  )).

%% --- Multiplies two integers.
%% (*)   :: Int -> Int -> Int
%% (*) external
user_tohnf('Prelude_*'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F * G),
		      freeze(B, true)
		  )).

%% --- Integer division. The value is the integer quotient of its arguments
%% --- and always truncated towards zero.
%% --- Thus, the value of <code>13 `div` 5</code> is <code>2</code>,
%% --- and the value of <code>-15 `div` 4</code> is <code>-3</code>.
%% div   :: Int -> Int -> Int
%% div external
user_tohnf('Prelude_div'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F // G),
		      freeze(B, true)
		  )
	      ).
%% --- Integer remainder. The value is the remainder of the integer division and
%% --- it obeys the rule <code>x `mod` y = x - y * (x `div` y)</code>.
%% --- Thus, the value of <code>13 `mod` 5</code> is <code>3</code>,
%% --- and the value of <code>-15 `mod` 4</code> is <code>-3</code>.
%% mod   :: Int -> Int -> Int
%% mod external
user_tohnf('Prelude_mod'(D, E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F mod G),
		      freeze(B, true)
		  )
	      ).

%% Comparison operators.

%% --- Is one integer less than another?
%% (<)   :: Int -> Int -> Bool
%% (<) external
user_tohnf('Prelude_<'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F < G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

%% --- Is one integer less than another?
%% (>)   :: Int -> Int -> Bool
%% (>) external
user_tohnf('Prelude_>'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F > G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

%% --- Is one integer greater than or equal to another?
%% (>=)  :: Int -> Int -> Bool
%% (>=) external
user_tohnf('Prelude_>='(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F >= G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

%% --- Is one integer less than or equal to another?
%% (<=)  :: Int -> Int -> Bool
%% (<=) external
user_tohnf('Prelude_<='(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F =< G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Igualdad
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% --- The always satisfiable constraint.
%% success :: Success
%% success external
user_tohnf('Prelude_success'#'Prelude_Success', 'Prelude_Success').

%% --- The equational constraint.
%% --- (e1 =:= e2) is satisfiable if both sides e1 and e2 can be
%% --- reduced to a unifiable data term (i.e., a term without defined
%% --- function symbols).
%% (=:=)   :: a -> a -> Success
%% (=:=) external
user_tohnf('Prelude_=:='(A, B)#'Prelude_Success', 'Prelude_Success') :-
	tohnf(A, A1),
	tohnf(B, B1),
	eq(A1, B1).

%% --- General equality builtin
%% (==) :: a -> a -> Bool
%% (==) external
user_tohnf('Prelude_=='(A,B)#C,C) :- 
	tohnf(A,A1),
	tohnf(B,B1),
	(
	    freeze(A1, (
			   freeze(B1, (
					  '=='(A1,B1,C)
				      )
				 )
		       )
		  )
	).
% 	(
% 	    eq(A1, B1) ->
% 	    C = 'Prelude_True'
% 	;
% 	    C = 'Prelude_False'
% 	).

'=='(A, B, 'Prelude_True') :-
	eq(A, B).
%	!.

'=='(A, B, 'Prelude_False') :-
	\+ eq(A, B).

%% --- Concurrent conjunction on constraints.
%% --- An expression like (c1 & c2) is evaluated by evaluating
%% --- the constraints c1 and c2 in a concurrent manner.
%% (&)     :: Success -> Success -> Success
%% (&) external
user_tohnf('Prelude_&'(A, B)#C, C) :-
	%% Will be concurrent in the future.
	tohnf(A, R1),
	R1 = 'Prelude_Success',
	tohnf(B, R2),
	R2 = 'Prelude_Success',
	C = 'Prelude_Success'.

%% --- Constrained expression.
%% --- An expression like (c &> e) is evaluated by first solving
%% --- constraint c and then evaluating e.
%% (&>)          :: Success -> a -> a
%% (&>) external
user_tohnf('Prelude_&>'(Constraint, Expr)#R, R) :-
	tohnf(Constraint, CR),
	CR = 'Prelude_Success',
	tohnf(Expr, R).

user_tohnf('Prelude_tonf'(Expr)#R, R) :-
	tonf(Expr, R).

%% Nonvar primitive, it doens't work as now but it is a start.
user_tohnf('Prelude_nonvar'(A)#A, A) :-
	freeze(A, true). % when?


%% FIXME: Rewrite this mess. Lots of bugs have arisen from bad cuts
%% here.
%% Equality with bindings.
eq(A,B) :-
	var(A),
	!,
	tonf(B,A).
%% 	!.

eq(A,B) :-
	var(B),
	!,
	tonf(A,B).
%%	!.
%% Those two cuts are wrong by the same reasons as the one below.

%% Fixme: 
%%% Question: Is this ok? 
%%% Answer: Umm, we should be careful with the posible unification
%%% we'd get. For instance, if we write: 
%%%  > map id == map id 
%%% That returns false, as the vars are surely different.
eq(A, A) :-
	A = pap(_, _, _)#_,
	!.

eq([],[]) :- 
	!.
eq([H|Ts], [A|Zs]) :- 
	!,
	tohnf(H,H1),
	tohnf(A,A1),
	eq(H1,A1),
	tohnf(Ts,Ts1),
	tohnf(Zs,Zs1),
	eq(Ts1,Zs1).
%	!.
%% This cut is not necessary, it makes 
%% let x, y free in x ++ y =:= [1,2,3,4] only return the 2 first sols.

%% Ugh!
eq(A,B) :-
  A =.. [Ca | Alist],
  B =.. [Cb | Blist],
  Ca = Cb,
  eq(Alist, Blist).

%eq(A, A) :-
%	!.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

