%% Algebraic diseq support for Curry.
%%
%% Copyright (C) 2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module('Diseqlib', [], [assertions]).

:- use_module('Core', [tonf/2, tohnf/2]).

:- multifile     user_tohnf/2, user_tonf/2, cons_meta/3.
:- discontiguous user_tohnf/2, user_tonf/2, cons_meta/3.

%% We are going to use fd.
:- use_package(fd).

%% Hard hard :)
:- use_module(library(write), [write/1, write/2]).
:- use_module(library(lists), [append/3, length/2]).

%% Integer operations.
user_tohnf('Diseq_=/='(D, E)#B, B) :- 
	diseq(D, E),
	B = 'Prelude_Success'.

%% Disequality support. 

%% Our constraint store is just: 

%% st(Var, Type, Terms) where Terms is an fd variable constrained
%% representing the forbidden constructors. Var is the variable to
%% which belongs the attr.

%% We need a 4th parameter due to bugs in Ciao fd library. (Which I
%% hope to fix soon

:- multifile verify_attribute/2.
:- multifile combine_attributes/2.

diseq(An, Bn) :-
	tohnf(An, A),
	tohnf(Bn, B),
	!, % This seems a sensible thing to do? May be wrong with non-det functions.
	diseq_trace('Diseq between'(A, B)),
	diseq_trace_info(A),
	diseq_trace_info(A),
	(
	    var(A), var(B) ->
	    diseq_vars(A, B)
	; % One of the parameters is not ground
	    var(A) ->
	    diseq_one_var(A, B)
	;
	    var(B) ->
	    diseq_one_var(B, A)
	; % both are instantiated.
	    diseq_spine(A, B)
	).

diseq_vars(A, B) :-
	AFd in 1..20, %Arbitrary to fix Ciao bugs.
	BFd in 1..20, %Arbitrary to fix Ciao bugs.
	AFd .<>. BFd,
	add_to_store(A, T, AFd, [(B, BFd)]),
	add_to_store(B, T, BFd, [(A, AFd)]).

%% Terms is in HNF.
diseq_one_var(Var, Term) :-
	get_cons(Term, Cons),
 	cons_meta(Cons, Type, (Index, Total)),
	(
	    %% The first case is to force the cons to be different.
	    N in 1..Total, N .<>. Index,
	    add_to_store(Var, Type, N, [])
	;   
	    cons_new_term(Term, Var),
	    diseq_spine(Var, Term)
	).
	    
diseq_spine(Term1, Term2) :-
	Term1 =.. [C1|L1],
	Term2 =.. [C2|L2],
	(
	    \+ C1 = C2 ->
	    true
	;
	    C1 = C2,
	    diseq_list(L1, L2)
	).
	    
cons_new_term(Term, Var) :-
	functor(Term, Cons, Arity),
	length(L, Arity),
	Var =.. [Cons|L].

diseq_list([A|AL],[B|BL]) :-
	diseq_trace(diseq_list(A, B)),
	(
	    diseq(A, B)
	;
	    diseq_list(AL, BL)
	).

%% This is called when we try to combine the info.
verify_attribute('$de_store'(Var, Type, Fd, List), Term) :-
 	diseq_trace('Checking attr'((Var, Fd), Term)),
 	diseq_trace_info(Var),
	get_cons(Term, Cons),
  	cons_meta(Cons, Type, (Index, Total)),
	Fd in 1..Total, %% The Fd var can be unbound when we arrive here.
	update_ranges(Total, List),
	Fd = Index, % This should propagate to the binded vars.
 	check_consistency_fd(Fd),
	detach_attribute(Var), 
	Var = Term,
	update_ranges(100, List).

combine_attributes('$de_store'(V1, Type1, Fd1, _L1), '$de_store'(V2, Type2, Fd1, _L2)) :-
	diseq_trace('Joining stores'(V1, Fd1, Type1, V2, Fd2, Type2)),
	Fd1 = Fd2,
	Type1 = Type2, %% Fixme, should update ranges.
	check_consistency_fd(Fd1),
	check_consistency_fd(Fd2).

add_to_store(Var, Type, Fd, L) :-
	(
	    get_attribute(Var, '$de_store'(Var, Type, FdOld, LOld)) ->
	    append(L, LOld, LNew),
	    Fd = FdOld,
	    check_consistency_fd(Fd),
	    update_attribute(Var, '$de_store'(Var, Type, Fd, LNew))
	;
	    attach_attribute(Var, '$de_store'(Var, Type, Fd, L))
	).

check_consistency_fd(N) :-
 	(
	    copy_term(N, N1),
	    labeling([N1]),
	    !,
	    diseq_trace('Var fd is consistent'(N))
	;
	    diseq_trace('Var fd is not consistent'(N)),
	    fail
	).

update_ranges(_Total, []).
update_ranges(Total, [(V,Fd)|LL]) :-
 	diseq_trace_info(V),
	Fd in 1..Total,
	update_ranges(Total, LL).

%% Aux
get_cons(Term, Cons) :-
	functor(Term, Cons, _).

%% Attribute output.
portray_attribute('$de_store'(_Var, _Type, _Fd, _), _Var) :-
	write(hola).

%% Debug
diseq_trace_info(_) :- !.
diseq_trace_info(A) :-
	write(user_error, 'Info for var'(A)),
	(
	    get_attribute(A, '$de_store'(_Var, Type, Fd, L)) ->
	    write(var_type(Type)),
	    write(fd_var(Fd)),
	    write(fd_var(L))
	;
	    write(is_free(A))
	),
	nl.

diseq_trace(_) :- !.
diseq_trace(X) :-
	write(user_error, X),
	nl.
