%% Sloth's compilation manager.
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(comp_manager, [compileFile/2, compileFilePrelude/2, 
			 rewrite/1, set_debug_paths/0]).

% Library modules:
:- use_module(library(dec10_io), [tell/1, told/0] ).
:- use_module(library(terms),    [atom_concat/2] ).
:- use_module(library(io_alias_redirection)).


% Sloth modules:
:- use_module(symtab,     [new_symtab/1, delete_symtab/1] ).
:- use_module(scope,      [new_scope/1, delete_scope/1] ).
:- use_module(cmdopts,    [cmd_options/2, set_option/2, unset_option/1] ).
:- use_module(debug,      [dump_abstree/2, dump_deftrees/2, dump_rules/2,
	                   pretty_list_write/1, diff_lists/3, debug_pass/1,
			   debug_write/1]).

:- use_module(parser,     [parse_module/2] ).
:- use_module(unref,      [solve_all_refs/1, free_all_vars/1, 
	                   solve_all_type_refs/1, free_all_type_vars/1]).
:- use_module(lambda,     [desugar_lambdas/1] ).
:- use_module(localdefs,  [desugar_ldef/1] ).
:- use_module(trans_lazy, [trans_module/1]).
:- use_module(typer,      [type_module/1] ).
:- use_module(exports,    [fix_exports/1] ).
%:- use_module(verifier,   [verify_rules/1] ).

:- use_package(hiord).

% Compiler passes. Be aware that they're order depending.
passes([solve_all_refs, desugar_ldef, desugar_lambdas, free_all_vars,
	solve_all_type_refs, free_all_type_vars, fix_exports]).

%% Main compilation function.
compileFile(In, Out) :-
	catch(compileFileReal(In, Out), E, do_comp_error(E, In)).

do_comp_error(E, In) :-
	comp_manager_cleanup(In),
	throw(E).

comp_manager_setup(In) :-
	(
	    comp_manager_cleanup(In) 
	;
	    true
	),	    
	new_symtab(In),
	new_scope(In).

comp_manager_cleanup(In) :-
	delete_scope(In),
	delete_symtab(In).

compileFileReal(In, Out) :-
	comp_manager_setup(In),
	debug_pass("Start"), 
	parse_module(In, Module),
	debug_pass("Parser"), 
	!,
	start_debug(Out, Module),
	rewrite(Module),
	!,
	type_module(Module),
	debug_pass("Typer"), 
	stop_debug,
	!,
	tell(Out),
	trans_module(Module),
	told,
	!,
	debug_pass("Translation"),
	comp_manager_cleanup(In),
	debug_pass("End").

compileFilePrelude(In, Out) :-
	set_option('-np', none),
	compileFile(In, Out),
	unset_option('-np').

rewrite(Module) :-
	passes(P),
	do_passes(Module, P).

do_passes(_, []).
do_passes(M, [P|PR]) :-
%	write(starting_pass(P)),
	P(M),
	atom_codes(P, PS),
	debug_pass(PS),
	do_debug(M, P),
	!,
	do_passes(M, PR).

:- dynamic last_data/1.

start_debug(Module, File) :-
	cmd_options('-v', _),
	atom_concat([File, '.debug'], FileDebug),
	%% Should set the debug handler here.
	tell(FileDebug),
	dump_abstree(Module, Abst),
	pretty_list_write(Abst),
	assert(last_data(Abst)),
	!.

start_debug(_,_) :- !.

stop_debug :-
	cmd_options('-v', _),
	retract(last_data(_)),
	dump_deftrees(Module, Dt),
	pretty_list_write(Dt),
	dump_rules(Module, Abst),
	pretty_list_write(Abst),
	told.

stop_debug :- !.

do_debug(Module, Phase) :-
	(
	    cmd_options('-v', _) ->
	    debug_write('\nPhase: '),
	    debug_write(Phase),
	    debug_write('\n'),
	    retract(last_data(OldAbs)),
	    dump_abstree(Module, NewAbst),
	    assert(last_data(NewAbst)),
	    diff_lists(OldAbs, NewAbst, DAbs),
	    pretty_list_write(DAbs)
	;
	    true
	).

set_debug_paths :-
	set_option('-l', ['../lib']).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

