%% Curry AST handling.
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(curryast, [split_rule/3, get_params/2, get_function/2,
	             param_at/3, map_tree/3, map_rules/2,
		     get_variables/2, get_identifiers/2, get_calls/2, 
		     check_tree/1, module_name/2, function_name/2, id_name/3,
		     map_tree_list/3]).

% We use high-order prolog.
:- use_package(hiord).

%%% Real includes.
:- use_module(library(terms), [atom_concat/2] ).
:- use_module(library(lists), [nth/3, list_concat/2] ).

:- use_module(symtab,         [get_rules/2, get_real_rules/2, 
	                       update_rules/2, get_object_property/3]).

%% In the future would be nice to export those from here, or maybe
%% providing enought predicates here such that no module needs to know the operators.
:- include(ast_ops).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Structural predicates.
%% Should be used when posible instead including ast_ops.

split_rule(Rule, Head, Body) :- 
	Rule = Head := Body.

%% Encapsulate qid
module_name(qid(M,_), M).
function_name(qid(_, F), F).

id_name(qid(M, F), M, F).

get_params(_@Params, Params) :- !.

%% TODO: Maybe we should handle constants.
get_function(F@_, F) :- !.
get_function(F, F) :- !.

param_at(FunApp, [], FunApp).
param_at(_@PList, [Pos|Poss], Exp) :-
	nth(Pos, PList, TExp),
	param_at(TExp, Poss, Exp).

%% Gets all the Expr variables (param ATM).
get_variables(Expr, Vars) :-
	map_tree(Expr, get_var(Vars), _),
	length(Vars, _), % Hacky, hacky.
	!.

%% For use with map_tree.
get_var(param(X), Dicc, const(_,_)) :-
	member(param(X), Dicc).

%% Gets all the expr's called functions.
get_identifiers(Expr, Vars) :-
	map_tree(Expr, get_iden(Vars), _),
	length(Vars, _), % Hacky, hacky. :)
	!.

%% For use with map_tree.
get_iden(qid(M, N), Dicc, const(_,_)) :-
	member(qid(M, N), Dicc).

%% get_calls(+Function, -Ids)
%% F calls Ids
get_calls(F, Ids) :-
	get_rules(F, Rules),
	(
	    Rules = [builtin(_)] ->
	    Ids = []
	;
	    all_rules_fun(Rules, RIds),
	    list_concat(RIds, Ids)
	).

all_rules_fun([], []).
all_rules_fun([R|RS], [Ids|IS]) :-
	split_rule(R, _Head, Body),
	get_identifiers(Body, AllIds),
	only_functions(AllIds, Ids),
	all_rules_fun(RS, IS).

%% Filters contructors and non-local functions.
only_functions([], []).
only_functions([F|FS], RS) :-
	(
	    get_object_property(F, constructs, _)
	;
	    get_rules(F, Rules),
	    Rules = [builtin(_)]
	),
	!,
	only_functions(FS, RS).

only_functions([F|FS], [F|RS]) :-
	!,
	only_functions(FS, RS).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Map over AST first attempt.
%% This needs some care because our AST is not an algebraic data type.

%% HOW TO USE??
%% You must supply a predicate that fails without doing anything if you 
%% don't want to operate on this term, or succeeds giving the transformed
%% term.

:- meta_predicate map_tree(?, pred(2), ?).

%% Map tree works by trying to apply first the user supplied
%% predicate. If that predicate fails (ie. not applicable to the
%% current term) we'll backtrack and default rules will be applied.
map_tree(E, Func, NE) :-
	Func(E, E1),  % This fails depending on the specified Func.
	!,
	(
	    E == E1 ->
	    E1 = NE   % Or maybe fail, so we continue with other transformations.
	;
	    map_tree(E1, Func, NE)
	).

%% If we have no match, recurse structurally over the AST.
map_tree(const(A, B), _F, const(A, B)) :-
	!.
map_tree(freevar(A), _F, freevar(A)) :-
	!.
map_tree(param(A), _F, param(A)) :-
	!.
map_tree(qid(M, N), _F, qid(M, N)) :-
	!.
map_tree(gunref(X), _F, gunref(X)) :-
	!.
map_tree(E@EL, Func, NE@NEL) :-
	map_tree(E, Func, NE),
	!,
	map_tree_list(EL, Func, NEL).

map_tree(lambda(Pattern, Expr), Func, lambda(NPattern, NExpr)) :-
	map_tree(Pattern, Func, NPattern),
	!,
	map_tree(Expr, Func, NExpr).

map_tree(ldef(LDefs, Expr), Func, ldef(LDefs2, NExpr)) :-
	map_tree_list(LDefs, Func, LDefs2),
	!,
	map_tree(Expr, Func, NExpr).

%% For ldefs
map_tree(pattern(Pat, Expr), F, pattern(Pat1, Expr1)) :-
	map_tree(Pat, F, Pat1),
	map_tree(Expr, F, Expr1).

map_tree(Head := Body, F, Head1 := Body1) :-
	map_tree(Head, F, Head1),
	map_tree(Body, F, Body1).

%% This is needed because we don't know the type.
map_tree(untyped_cond(E), F, untyped_cond(ET)) :-
	map_tree(E, F, ET).

%% Rules mapping.

:- meta_predicate map_tree_list(?, pred(2), ?).

map_tree_list([], _F, []) :- 
	!.

map_tree_list([E|EL], F, [NE|NEL]) :-
	map_tree(E, F, NE),
	!,
	map_tree_list(EL, F, NEL).

:- meta_predicate map_rules(?, pred(2)).

%% Fixme, use a real map.
map_rules(Module, Function) :-
	get_real_rules(Name, Rules),
	map_rules_list(Rules, Function, NRules),
	update_rules(Name, NRules),
	fail.
map_rules(_M, _F).

:- meta_predicate map_rules_list(?, pred(2), ?).

map_rules_list([], _F, []) :-
	!. %% This is here because our loop is side-effect and fail.

%% Ok, this needs some explanation: We are supposed to apply the
%% function to each rule separately, but with the current interface we
%% are getting the same function (that usually hold some state)
%% applied to all the rules. One of the side effects of this is that
%% we end using the wrong dictionary for rules.

%% Fortunately, due to Ciao hiord implementation we can use copy_term
%% to get a fresh new function ;)

map_rules_list([R|RL], Function, [NR|NRL]) :-
	copy_term(Function, UseAndThrowF),
	split_rule(R,  Head, Body),
	map_tree(Head, UseAndThrowF, NHead),
	map_tree(Body, UseAndThrowF, NBody),
	split_rule(NR, NHead, NBody),
	!,
	map_rules_list(RL, Function, NRL).

%% check_tree(AST), takes a Curry AST and verifies that is correct.

check_tree(builtin(_)).

check_tree(Head := Body) :-
	check_head(Head),
	check_expr(Body).

check_head(qid(_Mod, _Name)@Pattern) :-
	check_pattern(Pattern).

check_pattern(P) :-
	list(P),
	check_expr_list(P).

%check_expr(unref(_)) :-
%	write(warning).

check_expr(const(_, _)).
check_expr(freevar(_)).
check_expr(param(_)).
check_expr(qid(_,_)).

check_expr(lambda(Pattern, Expr)) :-
	check_expr(Pattern),
	!,
	check_expr(Expr).

check_expr(ldef(_LDefs, Expr)) :-
%	check_defs(...
	check_expr(Expr).

check_expr(Expr@ExprList) :-
	check_expr(Expr),
	check_expr_list(ExprList).

check_expr_list([]).
check_expr_list([E|Es]) :-
	check_expr(E),
	!,
	check_expr_list(Es).

check_func(qid(_Mod, _Name)).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

