%% Sloth's shell.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(shell, [start_shell/0, start_shell_debug/0]).

:- use_module(library(strings),    [get_line/1]).
:- use_module(library(lists),      [list_concat/2]).
:- use_module(library(format),     [format/2]).
:- use_module(library(compiler),   [unload/1]).
:- use_module(library(terms),      [atom_concat/2] ).
:- use_module(library(filenames),  [file_name_extension/3, 
                                    file_directory_base_name/3]).
%% Expression compiling.
:- use_module(parser,       [parse_expr/2]).
:- use_module(trans_expr,   [trans_expr/2] ).
:- use_module(curryast,     [map_tree/3] ).
:- use_module(unref,        [free_vars/3, free_all_vars/1, solve_refs/2] ).
:- use_module(localdefs,    [proccess_ldef_global/2, proccess_ldef_pattern/2, proccess_ldef_func/2]).
:- use_module(lambda,       [new_lambda/2]).
:- use_module(trans_lazy,   [trans_module_named/2]).
:- use_module(comp_manager, [rewrite/1, compileFile/2]).
:- use_module(scope,        [new_scope/1, delete_scope/1, 
	                     open_scope/1, close_named_scope/1]).
:- use_module(symtab,       [new_symtab/1, delete_symtab/1,
	                     set_object_property/3, delete_property/2,
			     get_module_functions/2, delete_rules/1]).

%% Module handling and loading.
:- use_module(modules,      [load_module/3, unload_module/1, 
                             get_loaded_modules/1, module_is_loaded/1]).
%% Misc.
:- use_module(cmdopts,      [cmd_options/2, set_option/2]).
:- use_module(typer,        [type_module/1, infer_type_expr/2]).
:- use_module(types,        [pretty_type/2]).
:- use_module(error,        [do_error/1]).

:- use_package(hiord).

%% Pretty printing.
:- use_module(shell_printing, [pre_printing/3, post_printing/3]).

%% Disabled ATM, this doesn't work in Ciao for global installations yet. :(
%:- use_module(readline,     [readline/2, add_history/1]).

start_shell :-
	reset,
	loop.

%% This is for calling within the CIAO debugger.
start_shell_debug :-
	working_directory(X, X),
	atom_concat([X, '/../lib'], CDir),
	set_option('-l', [CDir]),  
	start_shell.  

%% Shell main loop. Inspired by some Prolog code I don't exactly
%% remember, maybe the CIAO shell one.
loop :-
	repeat,
	sloth_prompt(P),
	atom_codes(P1, P),
	prompt(_, P1),
	%readline(P, Line1),
        %add_history(Line1),
	get_line(Line1),
	(
	    is_command(Line1, Command) ->
	    do_command(Command)
	;
	    append(Line1, [0'\n], Line), %% To be layout compliant.
	    open_scope(parser(shell_expr)), %% 
	    (
		prompt(_, ''),
		exec_expr(Line) ->
		close_all_parsing_scopes
	    ;
		close_all_parsing_scopes
	    )
	),
	Command == quit,
	cleanup,
	!.

:- meta_predicate map_apply(?, pred(1)).
map_apply([], _).
map_apply([I|L], F) :-
	F(I),
	!,
	map_apply(L, F).

%% This is little messy, see ticket #?
close_all_parsing_scopes :-
	close_named_scope(parser(_)),
	get_module_functions('__sloth_Shell', F),
	map_apply(F, delete_rules),
%%	write(deleted(F)),
	open_scope(parser('__sloth_Shell')),
	!. %% Needed :(

%% We've got an expression, execute it.
exec_expr(Input) :-
	!,
	catch(translate(Input, Dicc, Prolog, TmpMod, Type), E, do_shell_error(E)),
	pre_printing(Dicc, Prolog, PrintingInfo),
	%% At this point we've got the compilation right.
	!,
	Query = tonf(Prolog, Res),
	(
	    %% load the shell module (generated in translate.
	    use_module(TmpMod),
	    %% Ah! The beauty of Prolog, all this shell is just some
	    %% silly code and a call/1.
       	    statistics(runtime, [_,_]),
	    call(Query), 
       	    statistics(runtime, [_,Time2]),
	    format("Execution time: ~p ms\n", [Time2]),
	    post_printing(Res, Type, PrintingInfo),
	    try_more                 % Wanna get more?
	;
	    write('No solutions'),
	    nl
	),
	% unload the shell module
	!,
	unload(TmpMod),
	fail.

%% FIXME:
%% This is high-level compilation logic and should be refactored to 
%% comp_manager or somewhere else.

%% The idea is to have a compileExpression(OutPutModule, Expr, Prolog,
%% Type) in comp_manager.
translate(Input, Dicc, Prolog, TmpMod, Type) :-
	set_object_property('__sloth_Shell', exports, []),
	parse_expr(Input, Expr),
	!,
	(
	    map_tree(Expr, solve_refs, SExpr),
	    map_tree(SExpr, proccess_ldef_func, FExpr),
	    map_tree(FExpr, proccess_ldef_global, PExpr),
	    map_tree(PExpr, proccess_ldef_pattern, LExpr),
	    map_tree(LExpr, new_lambda, TExpr),
	    map_tree(TExpr, free_vars(Dicc), ExprFree),
	    free_all_vars('__sloth_Shell'),
	    %% Code output
	    gen_shell(TmpMod),
	    !,
	    % Generate the shell module, any function generated during
	    % the compilation proccess will go to that temp module.
	    % Typing.
	    copy_term(ExprFree, ExprToType), %% Argggh.
	    infer_type_expr(ExprToType, Type),
	    !,
	    trans_expr(ExprFree, Prolog),
	    delete_property('__sloth_Shell', exports),
	    !
	;
	    delete_property('__sloth_Shell', exports),
	    throw(error('Failed compilation (We couldn\'t identify the error)'))
	).

%% Weird hack. This will be killed in version 2.0, around year 
%% 2016 ;)
gen_shell(TmpMod) :-
	rewrite('__sloth_Shell'),
	type_module('__sloth_Shell'),
	%% Output the code to a tmp file.
	mktemp('/tmp/__sloth_ShellXXXXXX', TmpMod),
	tell(TmpMod),
	file_directory_base_name(TmpMod, _, TmpFile),
	trans_module_named(TmpFile, '__sloth_Shell'),
	told.

%% I wish the Curry grammar would be like this one.
try_more :-
	write('Try more (y/n)? '),
	get_line(Line),
	\+ Line = [0'y|_].

%% "A Dios rogando y con el mazo dando"
%% On a second thought, is kinda funny that the reset routine is the
%% same that the initialization one.
reset :-
	(
	    delete_symtab(shell),
	    delete_scope(shell)
	;
	    true
	),
	new_symtab(shell),
	new_scope(shell),
	%% Doing something heavier destroys the shell itself.
	use_module(sloth('Core'), [tonf/2, tohnf/2]),
	load_module('Prelude', [], none),
	load_module('IO', [], none),
	load_module('Diseq', [], none),
	update_prompt,
	open_scope(parser('__sloth_Shell')).

%% When we exit the shell, it's possible that we've got to do some
%% cleanup.
cleanup :-
%% I've heard some nasty things about using that rm command in beta
%% software, so better let the cruft in its place for now.
%%	rm('/tmp/__sloth_Shell.pl'),
	!.

%% This is an example of intelligent error handling.
do_shell_error(E) :-
	(
	    do_error(E),
	    shell_error_cleanup,
	    !,
	    fail
	;
	    write(E),
	    nl,
	    fail
	).

shell_error_cleanup :-
	close_all_parsing_scopes,
	(
	    delete_property('__sloth_Shell', exports)
	;
	    true
	).

%% 
compile_curry(Module) :-
	(
  	    file_name_extension(File, Module, '.curry'),
  	    file_exists(File) ->
 	    file_name_extension(PFile, Module, '.pl'),
 	    compileFile(File, PFile),
	    !
	;
	    true %% Umm, this should be cleaned up too.
	).

%% This needs more cleanup.
shell_load_module(Module) :-
	(
	    \+ module_is_loaded(Module) ->
	    catch(
		     (
			 compile_curry(Module),
			 load_module(Module, [], none)
		     ), E, do_shell_error(E)
		 ),		      
            update_prompt
        ;
	    true
	).

shell_unload_module(Module) :-
	(
	    unload_module(Module),
	    update_prompt, !
	;
	    format("Module ~p is not loaded.\n", [Module]), fail
	).

%% FIXME: This is not quite rigth.
:- data sloth_prompt/1.
update_prompt :-
	get_loaded_modules(ML),
	build_prompt(ML, Pr),
	(
	    retract(sloth_prompt(_))
	;
	    true
	),
	assert(sloth_prompt(Pr)),
	!.
	
build_prompt(ML, Pr) :-
	interleave_spaces(ML, ML1),
	list_concat(ML1, ML2),
	list_concat([ML2, "> "], Pr).

interleave_spaces([], []).
interleave_spaces([M], [M1]) :-
	!,
	atom_codes(M, M1). 
interleave_spaces([M|MS], [M1," "|MS1]) :-
	atom_codes(M, M1),
	interleave_spaces(MS, MS1).

%% Command parsing related.
%% Fixme: Question: What about a command parser?
%%        Answer: You'll see one the day CIAO has a readline binding.
is_command([0':|Command], Comm) :-
	command(Command, Comm).
is_command('', quit).
	
command("quit", quit).
command("q", quit).
command(Comm, load(Module)) :-
	append("load ", Module, Comm).
command(Comm, load(Module)) :-
	append("l ", Module, Comm).
command(Comm, unload(Module)) :-
	append("unload ", Module, Comm).
command(Comm, reload(Module)) :-
	append("reload ", Module, Comm).
command(Comm, type(Object)) :-
	append("type ", Object, Comm).
command(Comm, type(Object)) :-
	append("t ", Object, Comm).
command("reset", reset).

%%	
do_command(quit).
do_command(load(Module)) :-
	atom_codes(AModule, Module),
	!,
	shell_load_module(AModule).

do_command(unload(Module)) :-
	atom_codes(AModule, Module),
	!,
	shell_unload_module(AModule).

do_command(reload(Module)) :-
	atom_codes(AModule, Module),
	!,
	shell_unload_module(AModule),
	shell_load_module(AModule).
do_command(reset) :-
	reset.

%% FIXME: This needs cleanup
do_command(type(Object)) :-
	!,
	catch(translate(Object, _D, _P, _M, Type), E, do_shell_error(E) ),
	pretty_type(Type, TypeP),
	prettyvars(TypeP),
	print(TypeP),
	nl.

%% Bad bad hack. The correct solution is outlined in the CIAO manual
%% about write_term/1:
%
%         * *priority(*_prio_*):* _prio_ is an integer between 1 and
%           1200. If the term to be printed has higher priority than
%           _prio_, it will be printed parenthesized. Default value is
%           1200 (no term parenthesized).

:- multifile portray/1.
portray((A,B)) :- 
	write('('), 
	write(A), 
	write(','), 
	write(B),
	write(')').

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

