%% Sloth's shell pretty printing.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(shell_printing, [pre_printing/3, post_printing/3]).

:- use_module(library(terms_vars), [varset/2] ).
:- use_module(library(freeze),     [frozen/2]).

% Are those really needed?
:- use_package(clpq).
:- use_package(fd).

%
:- use_module(types, [pretty_type/2]).

%% This is really clean Prolog code. We have to build the set of
%% interesting variables present in the user-supplied expresion, so we
%% can display them later.
pre_printing(Dicc, Prolog, vi(Dicc, Binds)) :-
	varset(Prolog, Vs1),
	filter_bindings(Vs1, Dicc, Binds).

%% vi(Dicc, bindings) is the information about the variables. Dicc is
%% the dictionary built by the compiler and binds has that in an
%% internal format.
post_variables(vi(Dicc, Binds), Res) :-
	check_constraints(Binds, [], CProperties),
	diseq_tidy(CProperties, CPd),
% 	unbind(CPd, Dicc, CP),
% 	prettyvars(CP),
	Res = CPd,
	!.

post_variables(vi(Dicc, Binds), Res) :-
	check_constraints(Binds, [], CProperties),
	format("Fail in post_variables: [~p]\n", [CProperties]).

%% Output predicates.
%% This get Res as Prolog code, Type of the expression and 
post_printing(Res, Type, VarInfo) :-
	post_variables(VarInfo, Cinfo), % Print binding info.
	RRes = Res,
	%%	    unbind(Res, Dicc, RRes),
        call_print(TypeP, RRes),
	pretty_type(Type, TypeP),
	prettyvars(TypeP),
	write(' :: '), print(TypeP), nl,
	VarInfo = vi(_, Binds),
	write_bindings(Binds), nl,
	format("Constraint info: ~p\n", [Cinfo]),
	!.

post_printing(Res, Type, VarInfo) :-
	format("Goal succeded, but couldn't print the information:\n", []),
	format("--> Additional debugging info: [Res: ~p,Type: ~p, VI: ~p]", [Res, Type, VarInfo]).


:- multifile cons_meta/3.
diseq_tidy([], []).
diseq_tidy([E|El], Elt) :-
	(
	    E = attach_attribute(V, '$de_store'(_, _, _, _)) ->
	    get_attribute(V, '$de_store'(_, T, Fd, Vl)),
	    (
		\+ var(T) ->
		findall(C, (labeling([Fd]), cons_meta(C, T, (Fd,_))), LC),
		%% findall(Fd, labeling([Fd]), LC),
                Et = '=:='(V, LC),
		Elt = [Et|ETT],
		diseq_tidy(El, ETT)
	    ;
		Et = '=/='(V, Vl),
		Elt = [Et|ETT],
		diseq_tidy(El, ETT)
	    )
	;
	    diseq_tidy(El, Elt)
	).
	    

%% This is much better now than it used to be. Anyways, we could tidy
%% up R (from dump/3) a little bit more.
%% This predicate shouldn't fail.0
:- multifile dump_constraints/3.
check_constraints([], Acc, R) :- 
	!,
	dump_constraints(Acc, Acc, R).
check_constraints([binding(_N, X)|XS], Acc, Prop) :-
	check_constraints(XS, [X|Acc], Prop).

%% Dealing with variable naming.
%% Pretty print the bindings.
%% FIXME: We should obtain the case x = y.
write_bindings(B) :-
	write('{'),
	write_bindings2(B),
	write('}').

%% This is almost good, it just needs a little more effort to get it right.
write_bindings2([binding(Name, X) | R]) :-
	(
	    var(X) ->
		(
		    frozen(X, _Goal) ->
		    format("Computation frozen on ~p", [Name])
		;
		    write(Name),
		    write(' is '), 
		    write(X)
%		    write(' is free')
		)
	;
	    write(Name),
	    write(' = '),
	    call_print(a, X)
	),
	(
	    R = [] ->
	    true
	;
	    write(', '),
	    write_bindings2(R)
	).
write_bindings2([]).

%% Filter bindings obtains the original free vars present in the 
%% curry expression.
filter_bindings([], _, []).
filter_bindings([X|XS], Dicc, [binding(Name, X)|BXS]) :-
	bind_member(Name-X, Dicc),
	!,
	filter_bindings(XS, Dicc, BXS).
filter_bindings([_|L], Dicc, BL) :-
	filter_bindings(L, Dicc, BL).
	
bind_member(Name-X, [Name-BX|_]) :-
	X == BX,
	!.
bind_member(B, [_|Rest]) :-
	!,
	\+ var(Rest),
	bind_member(B, Rest).

%% Unbind expressions
unbind(Expr, Dicc, UExpr) :-
	var(Expr),
	!,
	(
	    bind_member(UExpr-Expr, Dicc),
	    !	
	;
	    Expr = UExpr
	).
unbind(Expr, _Dicc, Expr) :-
	functor(Expr, _F, 0),
	!.
unbind(Expr, Dicc, UExpr) :-
	Expr =.. EList,
	!,
	unbind_list(EList, Dicc, UList),
	UExpr =.. UList.

unbind_list([], _, []).
unbind_list([L|LS], D, [R|RS]) :-
	unbind(L, D, R),
	!,
	unbind_list(LS, D, RS).

%% Arrrrrghhhh. The googles, they do nothing!
%% Indeed, this should be fixed soon by our "type classes development
%% branch, soon to be merged"

call_print(Type, R) :-
	use_module(sloth('Core'), [tonf/2, tohnf/2]),
	copy_term(Type, T),
	(
	    var(T) ->
	    E = tonf('IO_print'(R)#_, _)
	;
	    T == ['Prelude.Char'] ->
	    E = tonf('IO_putStr'(R)#_, _)
	;
	    T = 'IO.IO'(_)->
	    E = true
	; % Actually this is never reached, because partial applied
	  % functions are not in HNF
	    T = (_ '->' _) ->
	    E = tonf('IO_print'("Cannot show functions")#_, _)
	;
	    E = tonf('IO_print'(R)#_, _)
	),
	call(E).

