%% The mythical symtab.
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(symtab, 
	[add_rule/2, get_rules/2,
	 get_real_rules/2, update_rules/2,
	 set_object_property/3, get_object_property/3,
	 get_module_functions/2, delete_rules/1,
	 add_constructor/2, get_constructors/2, replace_constructors/2,
	 delete_property/2, replace_object_property/3,
	 object_arity/2, new_type/1, replace_type/2,
	 is_type/1, new_symtab/1, unload_interfaces/1,
	 delete_symtab/1, get_module_types/2]).

% This module implements a function oriented symbol table.

:- use_module(library(dynamic),    [assert/1, assertz/1, retract/1, retractall/1]).
:- use_module(library(lists),      [append/3] ).
% 

:- use_module(types,  [ type_arity/2, type_name/2 ]).

%% Fixme, we should kill this.
:- use_module(curryast, [ module_name/2 ]).
% Private dynamics

:- dynamic symtabs/1.
:- dynamic func_rules/3.
:- dynamic object_properties/4.

% Maybe this will be needed:
% :- dynamic type_alias/2.
:- dynamic type_list/2.
:- dynamic type_cons/3.

new_symtab(Sn) :-
	asserta(symtabs(Sn)),
	!.

delete_symtab(Sn) :-
	retract(symtabs(Sn)),
	retractall(func_rules(Sn, _, _) ),
	retractall(object_properties(Sn, _, _, _) ),
	retractall(type_list(Sn, _)),
	retractall(type_cons(Sn, _, _) ),
	!.

% Rule handling predicates.
add_rule(Name, AbsTree) :-
	symtabs(Sn),
	!,
	(
	    retract(func_rules(Sn, Name, L)) ->
	    append(L, [AbsTree], NewL),
	    assertz(func_rules(Sn, Name, NewL))
	;
	    assertz(func_rules(Sn, Name, [AbsTree]))
	).
	    
% If we don't have any rules, add the first.
update_rules(Name, Rules) :-
	symtabs(Sn),
	!,
	retract(func_rules(Sn, Name, _)),
	assertz(func_rules(Sn, Name, Rules)).

get_rules(Name, ListOfAbsTrees) :-
	symtabs(Sn),
	!,
	func_rules(Sn, Name, ListOfAbsTrees).

delete_rules(Name) :-
	symtabs(Sn),
	!,
	retract(func_rules(Sn, Name, _)).
	
%% Not very clean.
get_real_rules(Name, LABS) :-
	symtabs(Sn),
	!,
	func_rules(Sn, Name, LABS),
	LABS \= [builtin(_)].

%% get_module_functions(-Module, +Funtions)
%% Should get all the functions belonging to Module.
get_module_functions(Module, Functions) :-
	findall(Name, (
			  get_rules(Name, _),
			  module_name(Name, Module)
		      ), 
		Functions
	       ).

get_module_types(Module, Types) :-
	findall((Type,Cons), 
        	(
		    is_type(Type),
		    copy_term(Type, T2), %% FIXME: Another example of why our symbol table is completely fucked-up.
		    type_name(Type, tid(Module, _)),
		    (
			get_constructors(T2, Cons) ->
			true
		    ;
			Cons = []
		    )
		), Types).

% General property handling.
set_object_property(Name, Property, Value) :-
	check_property(Property),
	symtabs(Sn),
	!,
	assert(object_properties(Sn, Name, Property, Value)).

get_object_property(Name, Property, Value) :-
	check_property(Property),
	symtabs(Sn),
	!,
	object_properties(Sn, Name, Property, Value).

delete_property(Name, Property) :-
	symtabs(Sn),
	!,
	retractall(object_properties(Sn, Name, Property, _)).

replace_object_property(Name, Property, Value) :-
	delete_property(Name, Property),
	set_object_property(Name, Property, Value).

% Function properties are well defined and symbol
% table enforces this rectriction.
is_property(type).

is_property(exports).
is_property(imports).

%% Operators priority.
is_property(priority).
is_property(assoc).

% FIXME: provisional until I see if can replace link in the whole Sloth.
is_property(constructs).

% If P is not a property throw and error.
check_property(P) :-
	is_property(P),
	!.

check_property(P) :-
	throw(error(symtab, invalid_property(P))).

%% object_arity returns the expected number or args that a cons or pred
%% should take.
object_arity(Object, Arity) :-
	get_object_property(Object, type, Type),
	type_arity(Type, Arity),
	!.

% Type database:

new_type(Type) :-
	symtabs(Sn),
	!,
	\+ type_list(Sn, Type),
	!,
	assert(type_list(Sn, Type)).

new_type(Type) :-
	throw(error(symtab, redefining(Type))).

is_type(Type) :-
	symtabs(Sn),
	!,
	type_list(Sn, Type).

%% This is wrong, hopefully will be solved by #15.
replace_type(Type, NewType) :-
	symtabs(Sn),
	!,
	retract(type_list(Sn, Type)),
	assert(type_list(Sn, NewType)).
 
% A data type in Curry usually has some constructors.
%
% Data type -> List of Constructors.
% Data type -> arity, etc...
% Constructor -> Arity, Type

add_constructor(Type, Cons) :-
	symtabs(Sn),
	!,
	(
	    retract(type_cons(Sn, Type, ConsL)),
	    assert(type_cons(Sn, Type, [Cons|ConsL]))
	;
	    assert(type_cons(Sn, Type, [Cons]))
	).

get_constructors(Type, ListOfCons) :-
	symtabs(Sn),
	!,
	type_cons(Sn, Type, ListOfCons).

replace_constructors(Type, ListOfCons) :-
	symtabs(Sn),
	!,
	copy_term(Type, T),
	retract(type_cons(Sn, T, _)),
	assert(type_cons(Sn, Type, ListOfCons)).
	
%% Fixme, this doesn't belong here and it's very ugly.
%% Kills all references of the symbol table to the module.
%% ugly
:- include(type_ops).
unload_interfaces(Module) :-
	symtabs(Sn),
	retractall(func_rules(Sn, qid(Module,_), _) ),
	retractall(object_properties(Sn, qid(Module,_), _, _) ),
	retractall(object_properties(Sn, tid(Module,_)<@_, _, _) ),
	retractall(type_cons(Sn, tid(Module, _)<@_, _) ),
	retractall(type_list(Sn, tid(Module, _)<@_)).


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

