%% Curry expressions to prolog translator.
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Julio Mariño Carballo
%%   * Jose María Rey
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(trans_expr, [trans_expr/2, parser2trad/2], [assertions]).

:- include(ast_ops).
:- include(type_ops). %% Arggg.

:- use_module(trans_tools,      [make_name/2] ).
:- use_module(symtab,           [get_rules/2, get_object_property/3, 
                                 object_arity/2]).
% ugh.
:- use_module(typer,            [infer_type_expr/2]).

%%
:- use_module(library(lists),   [append/3, length/2, reverse/2]).

%% Misc tools.
:- pred list_take(L, N, LL) : list * int * list # "Takes N elements from L.".
list_take(_L, 0, []).
list_take([L|LR], N, [L|RR]) :-
	N1 is N - 1,
	list_take(LR, N1, RR).

%% users:
% in tr_rule:
% 	trans_expr(Body, BodyTrad),
% 	parser2trad(Head, HeadTrad),
% also:
% 	parser2trad(Head, HeadTrad), % Translates the pattern in the head.
% 	add_arguments(HeadTrad, [R0], HeadTrad2),
% 	trans_expr(Body, BodyTrad),

%% new possible helper preds, change_head and change_body.
trans_expr(Ast, Prolog) :-
	trans_struc_goal(Ast, TExpr),
	memoize(TExpr, Prolog, _Dicc).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parser2trad(freevar(X), X) :-
	!.

parser2trad(param(X), X) :-
 	!.

parser2trad(const(_, X), X) :- 
	!.

parser2trad(qid(M, Id), P1) :-
	make_name(qid(M, Id), P1),
	!.

parser2trad(P@Ts,T) :-
	!,
	P = qid(_, _),
	make_name(P, P1),
	parser2trad_list(Ts,Ts2),
	T =.. [P1|Ts2].

parser2trad_list([], []).
parser2trad_list([T|Ts], [T2|Ts2]) :-
	parser2trad(T,T2),
	parser2trad_list(Ts,Ts2).

%% trans_struc_goal(-ParserExp, +PrologExp)
%% ParserExp is in a@[Args] form, while PrologExp is in
%% final prolog form.

trans_struc_goal(param(V)@[], V) :- !.
trans_struc_goal(param(V)@Pars, memo(apply_list(V, Pars2))) :-
	!,
	trans_struc_goal_list(Pars,Pars2).

%% This kind of application flattening has to improve.
trans_struc_goal((F@Arg1)@Arg2, Res) :-
	append(Arg1, Arg2, ArgN),
	!,
	trans_struc_goal(F@ArgN, Res).

%% Translation of partial applications (both for functions and cons)
trans_struc_goal(FId@Pars, pap(MI, Quedan, Pars2)) :-
	get_rules(FId, _Rules), % Is a function
	object_arity(FId, Arity),
	length(Pars,NumPars),
	NumPars < Arity,
	!,
	Quedan is Arity - NumPars,
	reverse(Pars, ParsRev),
	make_name(FId, MI),
	trans_struc_goal_list(ParsRev,Pars2).

trans_struc_goal(FId@Pars, pap(XX, Quedan,Pars2)):- %% Constr. currif.
	get_object_property(FId, constructs, _Type),
	make_name(FId, XX),
	object_arity(FId, Arity),
	length(Pars, NumPars),
	NumPars < Arity,
	!,
	Quedan is Arity - NumPars,
	reverse(Pars, ParsRev),
	trans_struc_goal_list(ParsRev,Pars2).

%% Next 3 are for single cons or functions.
trans_struc_goal(FId, X) :-
	get_rules(FId, _Rules),
	!,
	make_name(FId, XX),
	object_arity(FId, Arity),
	(
	    Arity == 0 -> 
	    X = memo(XX)
	; 
	    X = pap(XX, Arity, [])
 	).

trans_struc_goal(FId@[], X) :-
	!,
	trans_struc_goal(FId, X).

trans_struc_goal(FId, X) :-
	get_object_property(FId, constructs, _Type),
	make_name(FId, XX),
	object_arity(FId, Arity),
	(
	    Arity == 0 -> 
	    X = XX
	; 
	    X = pap(XX, Arity, [])
 	).

%% Generic translation. Both for cons and for fun.
trans_struc_goal(FId@Pars, T) :- 
	get_object_property(FId, constructs, _T),
	!,
	trans_struc_goal_list(Pars, Parst),
	make_name(FId, XX),
	T =.. [XX|Parst].

trans_struc_goal(FId@Pars, memo(T)) :- 
	object_arity(FId, Arity),
	length(Pars, Arity),
	!,
	trans_struc_goal_list(Pars, Parst),
	make_name(FId, XX),
	T =.. [XX|Parst].

%% This simulates an over-app.
trans_struc_goal(FId@Pars, T) :-
	object_arity(FId, ArityF),
	length(Pars, ArityP),
	ArityP > ArityF,
	!,
	list_take(Pars, ArityF, ParsC),
	trans_struc_goal(FId@ParsC, CompleteApp),
	append(ParsC, ParsR, Pars),
	trans_struc_goal_list(ParsR, ExtraArgs),
	T = memo(apply_list(CompleteApp, ExtraArgs)).

%% This is the most awkward rule, for translating cond exps we have to
%% look up the type, and call a different function depending on it.
%% Fixme, this should be cleanly implemented by a rewriting.
trans_struc_goal(untyped_cond(List), R) :-
	copy_term(List, List1),
	infer_type_expr(List1, Type),
	Type = tid('Prelude', [])<@[tid('Prelude', tuple)<@[CondT, _B]],
	!,
	(
	    CondT = tid('Prelude', 'Bool')<@[] ->
	    Expr = qid('Prelude', bool_cond)@[List],
	    trans_struc_goal(Expr, R)
	;
	    CondT = tid('Prelude', 'Success')<@[] ->
	    Expr = qid('Prelude', success_cond)@[List],
	    trans_struc_goal(Expr, R)
	;
	    throw(error(translator, not_allowed_type_for_cond(CondT)))
	).

%% All the simple cases
trans_struc_goal(E, E2):-
	parser2trad(E,E2). 
  
trans_struc_goal_list([],[]).
trans_struc_goal_list([A|As], [B|Bs]) :-
	trans_struc_goal(A,B),
	trans_struc_goal_list(As,Bs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Memoization (include here as it is not big enougth to live in its own mod.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Just local memo :(
memberM(X, [Y|V]) :-
	var(V),
	var(Y),
	X = Y,
	!.

memberM(X#Res, [Y#Res|_]) :-
	X == Y.
memberM(X, [_|XS]) :-
	memberM(X, XS).

memoize(V, V, _Dicc) :-
	var(V).

memoize(memo(Exp), Exp1#Res, Dicc) :-
	!,
	memoize(Exp, Exp1, Dicc),
	memberM(Exp1#Res, Dicc).	

memoize(CExp, NExp, Dicc) :-
	CExp =.. [F|Args],
	memoize_list(Args, NArgs, Dicc),
	NExp =.. [F|NArgs].

memoize_list([], [], _Dicc).
memoize_list([R|RS], [S|SS], Dicc) :-
	memoize(R, S, Dicc),
	!,
	memoize_list(RS, SS, Dicc).
